from getpass import getpass

import bitshares_signing as bts

print("\033c")

user = input("Username: ")
wif = getpass("WIF: ")

order = bts.prototype_order(
    {
        "wif": wif,
        "account_name": user,
        "asset_name": "BTS",
        "currency_name": "HONEST.MONEY",
    },
)


order1 = order.copy()
order1["edicts"] = [{"op": "login"}]


order2 = order.copy()
order2["edicts"] = [
    {"op": "buy", "amount": 1, "price": 1},
    {"op": "sell", "amount": 1, "price": 1},
]


order3 = order.copy()
# add edicts
order3["edicts"] = [
    # cancel - cancel all `ids` or all open orders if `ids` is ["1.7.X"]
    {
        "op": "cancel",
        "ids": ["1.7.X"],
    },
]

order4 = order.copy()
order4["edicts"] = [
    # swap - selling `amount` of `currency` for `assets` at `price` on `pool`
    {
        "op": "swap",
        "amount": 1,
        "price": 1,
        "pool": "1.19.305",
    },
    # deposit_pool - deposit `amount_a` of `asset` and `amount_b` of `currency` into `pool`
    {
        "op": "deposit_pool",
        "amount_a": 1,  # asset
        "amount_b": 1,  # currency
        "pool": "1.19.305",
    },
    # transfer - move `amount` of `asset` to `account_id`
    {
        "op": "transfer",
        "amount": 1,
        "account_id": "1.2.581357",
    },
    # issue - issue `amount` of `asset` to `account_id`
    {
        "op": "issue",
        "amount": 1,
        "account_id": "1.2.581357",
    },
    # reserve - reserve `amount` of `asset`
    {
        "op": "reserve",
        "amount": 1,
    },
    # call - update a call order with a change in debt (`debt_delta`), change in
    #        collateral (`collateral_delta`), and target collateral ratio (`tcr`)
    {
        "op": "call",
        "debt_delta": 1,
        "collateral_delta": 1,
        "tcr": 0,
    },
    # fee_pool - claim `amount_to_claim` from the fee pool of `asset_id`, issued by
    #            `issuer`
    {
        "op": "fee_pool",
        "amount_to_claim": 1,
        "asset_id": "1.3.305",
        "issuer": "1.2.581357",
    },
    # create_pool - create a liquidity pool, where asset a and b are the asset and
    #               currency in the header.  The pool token is given by `share_asset`,
    #               and withdrawal and taker fees are given by their respective keys.
    {
        "op": "create_pool",
        "share_asset": "1.3.9999",
        "taker_fee_percent": 0.1,  # these two are in percent, so 0.1 would be amt*0.001
        "withdrawal_fee_percent": 0.1,
    },
    # create_asset - create a new asset, inline comments given because there are
    #                a lot of values.
]


order5 = order.copy()
order5["edicts"] = [
    {
        "op": "create_asset",
        "issuer": "1.2.581357",  # who is allowed to issue this asset?
        "symbol": "TEST",  # what is it called?
        "precision": 8,  # how many decimal points can amounts of it be referenced with?
        "common_options": {
            # how much of it can exist?
            "max_supply": 1000000,
            # the fee paid by the buyer to the issuer during limit orders
            "market_fee_percent": 1,
            # non-changeable number that limits the above value
            "max_market_fee": 2,
            # series of flags which limits the ability of the issuer to continue to edit particular flags
            "issuer_permissions": 0,
            # the ability of the issuer to change capabilities of the UIA such as the ability to transfer
            "flags": 0,
            # the user may pay their market fees using core token at the core exchange rate
            # periodically updated by the issuer from the fee pool maintained by the issuer
            "core_exchange_rate": {
                "base": {"amount": 1, "asset_id": "1.3.1"},
                "quote": {"amount": 1, "asset_id": "1.3.0"},
            },
            # the issuer may create a list of only those who have permission to trade this asset
            "whitelist_authorities": [],
            # as well as those who do not
            "blacklist_authorities": [],
            # the issuer may create a list of only those other assets this one may be traded against
            "whitelist_markets": [],
            # as well as those it cannot be
            "blacklist_markets": [],
            "description": "Asset for testing bitshares_signing",
        },
        "is_prediction_market": False,
    },
]

bts.broker(order1, broadcast=False)
bts.broker(order2, broadcast=False)
bts.broker(order3, broadcast=False)
bts.broker(order4, broadcast=False)
bts.broker(order5, broadcast=False)



r"""
build_transaction.py

  ____  _ _   ____  _                         
 | __ )(_) |_/ ___|| |__   __ _ _ __ ___  ___ 
 |  _ \| | __\___ \| '_ \ / _` | '__/ _ \/ __|
 | |_) | | |_ ___) | | | | (_| | | |  __/\__ \
 |____/|_|\__|____/|_| |_|\__,_|_|  \___||___/
       ____  _             _                  
      / ___|(_) __ _ _ __ (_)_ __   __ _      
      \___ \| |/ _` | '_ \| | '_ \ / _` |     
       ___) | | (_| | | | | | | | | (_| |     
      |____/|_|\__, |_| |_|_|_| |_|\__, |     
               |___/               |___/      


WTFPL litepresence.com Dec 2021 & squidKid-deluxe Jan 2024

Build graphene transactions from human-readable edicts

"""

import itertools
# STANDARD PYTHON MODULES
import time
from binascii import unhexlify  # hexidecimal to binary text
from struct import unpack_from  # convert back to PY variable

# GRAPHENE SIGNING MODULES
from .config import AUTOSCALE, CORE_FEES, DUST, KILL_OR_FILL, LIMIT
from .graphenize.asset_create import graphenize_asset_create
from .graphenize.call_order_update import graphenize_call
from .graphenize.fee_pool import graphenize_fee_pool
from .graphenize.issue_reserve import graphenize_issue, graphenize_reserve
from .graphenize.limit_orders import (graphenize_cancel,
                                      graphenize_limit_orders,
                                      scale_limit_orders)
from .graphenize.liquidity_pools import (graphenize_pool_creation,
                                         graphenize_pool_deposit,
                                         graphenize_swaps,
                                         graphenize_pool_update, graphenize_pool_delete)
from .graphenize.price_feeds import graphenize_add_producer, graphenize_publish
from .graphenize.transfer import graphenize_transfer
from .rpc import (rpc_account_id, rpc_balances, rpc_block_number,
                  rpc_lookup_asset_symbols, rpc_open_orders, rpc_tx_fees)
from .types import ObjectId
from .utilities import fraction, it, to_iso_date

# MAX is 4294967295; year 2106 due to 32 bit unsigned integer
END_OF_TIME = 4 * 10**9  # about 75 years in future
# very little
SATOSHI = 0.00000001
# almost 1
SIXSIG = 0.999999

DEV = True


def graphenize_login(login_edicts, fees, account_id, tx_operations):
    """
    Translate login to dummy graphene cancel order
    """
    if login_edicts:
        fee = {"amount": 0, "asset_id": "1.3.0"}
        operation = [
            2,
            {
                "fee": fee,
                "fee_paying_account": account_id,
                "order": "1.7.0",
                "extensions": [],
            },
        ]
        tx_operations.append(operation)
    return tx_operations


def build_transaction(rpc, order):
    """
    # this performs incoming limit order api conversion
    # from human terms to graphene terms

    # humans speak:
     - account name, asset name, order number
     - decimal amounts, rounded is just fine
     - buy/sell/cancel
     - amount of assets
     - price in currency

    # graphene speaks:
     - 1.2.x, 1.3.x, 1.7.x
     - only in integers
     - create/cancel
     - min_to_receive/10^receiving_precision
     - amount_to_sell/10^selling_precision

    # build_transaction speaks:
     - list of buy/sell/cancel human terms edicts any order in
     - validated data request
     - autoscale amounts if out of budget
     - autoscale amounts if spending last bitshare
     - bundled cancel/buy/sell transactions out; cancel first
     - prevent inadvertent huge number of orders
     - do not place orders for dust amounts
    """
    # VALIDATE INCOMING DATA
    for key, expected_type in [("edicts", list), ("nodes", list), ("header", dict)]:
        if not isinstance(order[key], expected_type):
            raise ValueError(
                f"order parameter '{key}' must be of type {expected_type.__name__}:"
                f" {order[key]}; is {type(order[key]).__name__}"
            )

    # the location of the decimal place must be provided by order
    asset_precision = int(order["header"].get("asset_precision", 5))
    asset_id = str(order["header"].get("asset_id", "1.3.0"))

    account_name = str(order["header"]["account_name"])
    account_id = str(
        order["header"].get("account_id", rpc_account_id(rpc, account_name))
    )

    currency_precision = int(order["header"].get("currency_precision", 0))
    currency_id = str(order["header"].get("currency_id", 0))

    checks = [account_id, asset_id]

    # perform checks on currency for limit orders
    if order["edicts"][0]["op"] in ["buy", "sell"]:
        checks.append(currency_id)
        if not currency_precision or not currency_id:
            return -1
    # validate a.b.c identifiers of account id and asset ids
    for check in checks:
        ObjectId(check)
    # GATHER TRANSACTION HEADER DATA
    # fetch block data via websocket request
    block = rpc_block_number(rpc)
    ref_block_num = block["head_block_number"] & 0xFFFF
    ref_block_prefix = unpack_from("<I", unhexlify(block["head_block_id"]), 4)[0]
    # fetch limit order create and cancel fee via websocket request
    fees = rpc_tx_fees(rpc, account_id)
    # establish transaction expiration
    tx_expiration = to_iso_date(int(time.time() + 120))
    if DEV:
        print(ref_block_num, ref_block_prefix, tx_expiration)
    # initialize tx_operations list
    tx_operations = []

    # Sort incoming edicts by type
    edict_types = {
        "buy": [],
        "sell": [],
        "cancel": [],
        "swap": [],
        "create_pool": [],
        "delete_pool": [],
        "create_asset": [],
        "transfer": [],
        "reserve": [],
        "issue": [],
        "login": [],
        "call": [],
        "pool_deposit": [],
        "pool_update": [],
        "fee_pool": [],
        "publish": [],
        "add_producer": [],
        # TODO
        # add those from `shorting attack avenger`
    }

    for edict in order["edicts"]:
        if edict["op"] in edict_types:
            print(it("yellow", str({k: str(v) for k, v in edict.items()})))
            edict_types[edict["op"]].append(edict)

    # convert to decimal type

    for op_type in ("buy", "sell"):
        for idx, item in enumerate(edict_types[op_type]):
            edict_types[op_type][idx]["amount"] = item["amount"]
            edict_types[op_type][idx]["price"] = item["price"]

    graphenize_cancel(
        rpc, edict_types["cancel"], fees, order, account_name, account_id, tx_operations
    )
    if any([edict_types["sell"], edict_types["buy"]]):
        scaled_limit_orders = scale_limit_orders(
            rpc,
            order,
            asset_id,
            currency_id,
            account_name,
            edict_types["buy"],
            edict_types["sell"],
        )
        tx_operations = graphenize_limit_orders(
            scaled_limit_orders,
            fees,
            [asset_id, asset_precision],
            [currency_id, currency_precision],
            account_id,
            tx_operations,
        )

    # fmt: off
    # Define a list of operations with their corresponding function and parameters
    operations = [
        ("swap", graphenize_swaps, [[asset_id, asset_precision], [currency_id, currency_precision]]),
        ("create_pool", graphenize_pool_creation, [[asset_id, asset_precision], [currency_id, currency_precision]]),
        ("create_asset", graphenize_asset_create, [[asset_id, asset_precision], [currency_id, currency_precision]]),
        ("call", graphenize_call, [[asset_id, asset_precision], [currency_id, currency_precision]]),
        ("pool_update", graphenize_pool_update, [[asset_id, asset_precision], [currency_id, currency_precision]]),
        ("pool_deposit", graphenize_pool_deposit, [[asset_id, asset_precision], [currency_id, currency_precision]]),
        ("delete_pool", graphenize_pool_delete, [[asset_id, asset_precision], [currency_id, currency_precision]]),
        ("fee_pool", graphenize_fee_pool, [[asset_id, asset_precision], [currency_id, currency_precision]]),
        ("transfer", graphenize_transfer, [asset_id, asset_precision]),
        ("issue", graphenize_issue, [asset_id, asset_precision]),
        ("reserve", graphenize_reserve, [asset_id, asset_precision]),
        ("publish", graphenize_publish, [rpc]),
        ("add_producer", graphenize_add_producer, []),
        ("login", graphenize_login, []),
    ]
    # fmt: on

    # Iterate over the operations and apply them to tx_operations
    for name, func, params in operations:
        tx_operations = func(
            edict_types[name], fees, *params, account_id, tx_operations
        )

    # prevent inadvertent huge number of orders
    tx_operations = tx_operations[:LIMIT]
    # the trx is just a regular dictionary we will convert to json later
    # the operations themselves must still be an OrderedDict
    trx = {
        "ref_block_num": ref_block_num,
        "ref_block_prefix": ref_block_prefix,
        "expiration": tx_expiration,
        "operations": tx_operations,
        "signatures": [],
        "extensions": [],
    }
    return trx



"""
graphene_signing.py

Graphene types and signing function required to authenticate operation being
broadcast to BitShares

WTFPL litepresence.com Dec 2021 & squidKid-deluxe Jan 2024
"""

# Standard Python modules
from binascii import hexlify, unhexlify
from hashlib import new as hashlib_new
from hashlib import sha256

# Third party modules
import secp256k1  # Replaced ecdsa with secp256k1

# Graphene signing modules
from .config import PREFIX

# Global constants
BASE58 = b"123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
HEXDIGITS = "0123456789abcdefABCDEF"


class Base58:
    """Abstraction layer for base58 encoded strings and their hex/binary representation"""

    def __init__(self, data, prefix=PREFIX):
        self._prefix = prefix
        if all(c in HEXDIGITS for c in data):
            self._hex = data
        elif data[0] in ["5", "6"]:
            self._hex = base58_check_decode(data)
        elif data[0] in ["K", "L"]:
            self._hex = base58_check_decode(data)[:-2]
        elif data[: len(self._prefix)] == self._prefix:
            self._hex = gph_base58_check_decode(data[len(self._prefix) :])
        else:
            raise ValueError("Error loading Base58 object")

    def __format__(self, _format):
        if _format.upper() != PREFIX:
            raise ValueError(f"Format must be {PREFIX}")
        return _format.upper() + str(self)

    def __repr__(self):
        """Return hex string representation"""
        return self._hex

    def __str__(self):
        """Return base58 string representation"""
        return gph_base58_check_encode(self._hex)

    def __bytes__(self):
        """Return raw bytes representation"""
        return unhexlify(self._hex)


def base58_decode(base58_str):
    """Convert base58 string to hexadecimal"""
    base58_text = bytes(base58_str, "ascii")
    num = 0
    leading_zeroes_count = 0

    for byte in base58_text:
        num = num * 58 + BASE58.find(byte)
        if num == 0:
            leading_zeroes_count += 1

    res = bytearray()
    while num >= 256:
        div, mod = divmod(num, 256)
        res.insert(0, mod)
        num = div
    res.insert(0, num)

    return hexlify(bytearray(1) * leading_zeroes_count + res).decode("ascii")


def base58_encode(hexstring):
    """Convert hexadecimal string to base58"""
    byteseq = bytes(unhexlify(bytes(hexstring, "ascii")))
    num = 0
    leading_zeroes_count = 0

    for byte in byteseq:
        num = num * 256 + byte
        if num == 0:
            leading_zeroes_count += 1

    res = bytearray()
    while num >= 58:
        div, mod = divmod(num, 58)
        res.insert(0, BASE58[mod])
        num = div
    res.insert(0, BASE58[num])

    return (BASE58[0:1] * leading_zeroes_count + res).decode("ascii")


def ripemd160(string):
    """Generate 160-bit RIPEMD hash"""
    r160 = hashlib_new("ripemd160")
    r160.update(unhexlify(string))
    return r160.digest()


def double_sha256(string):
    """Generate double SHA-256 hash"""
    return sha256(sha256(unhexlify(string)).digest()).digest()


def base58_check_encode(version, payload):
    """Encode with base58 check"""
    string = f"{version:02x}{payload}"
    checksum = double_sha256(string)[:4]
    return base58_encode(string + hexlify(checksum).decode("ascii"))


def gph_base58_check_encode(string):
    """Graphene-specific base58 check encoding"""
    checksum = ripemd160(string)[:4]
    return base58_encode(string + hexlify(checksum).decode("ascii"))


def base58_check_decode(string):
    """Decode base58 with checksum verification"""
    string = unhexlify(base58_decode(string))
    dec = hexlify(string[:-4]).decode("ascii")
    checksum = double_sha256(dec)[:4]
    if string[-4:] != checksum:
        raise ValueError("Checksum verification failed")
    return dec[2:]


def gph_base58_check_decode(string):
    """Graphene-specific base58 check decoding"""
    string = unhexlify(base58_decode(string))
    dec = hexlify(string[:-4]).decode("ascii")
    checksum = ripemd160(dec)[:4]
    if string[-4:] != checksum:
        raise ValueError("Checksum verification failed")
    return dec


class Address:
    """Represents a blockchain address"""

    def __init__(self, address=None, pubkey=None, prefix=PREFIX):
        self.prefix = prefix
        self.pubkey = Base58(pubkey, prefix=prefix)
        self._address = address


class PublicKey:
    """Handles public key operations"""

    def __init__(self, pk, prefix=PREFIX):
        self.prefix = prefix
        self._pk = Base58(pk, prefix=prefix)
        self.address = Address(pubkey=pk, prefix=prefix)
        self.pubkey = self._pk
        # Create a secp256k1 PublicKey object for internal operations
        self._secp256k1_pubkey = secp256k1.PublicKey(bytes(self), raw=True)

    def compressed(self):
        """Generate compressed public key"""
        # Use secp256k1 to serialize as compressed key
        return self._secp256k1_pubkey.serialize(compressed=True).hex()

    def un_compressed(self):
        """Generate uncompressed public key"""
        # Use secp256k1 to serialize as uncompressed key
        return self._secp256k1_pubkey.serialize(compressed=False).hex()

    def point(self):
        """Get ECDSA point representation"""
        # Get the uncompressed serialization to extract x and y coordinates
        uncompressed = self._secp256k1_pubkey.serialize(compressed=False)
        # The first byte is 0x04 (uncompressed prefix), then 32 bytes x, then 32 bytes y
        x = int.from_bytes(uncompressed[1:33], 'big')
        y = int.from_bytes(uncompressed[33:65], 'big')
        # Return a tuple representing the point (x, y)
        return (x, y)

    def __repr__(self):
        """Return hex representation"""
        return repr(self._pk)

    def __format__(self, _format):
        """Format the public key"""
        return format(self._pk, _format)

    def __bytes__(self):
        """Return raw bytes (33 bytes for compressed)"""
        return bytes(self._pk)


class PrivateKey:
    """Handles private key operations and public key derivation"""

    def __init__(self, wif=None, prefix=PREFIX):
        self._wif = wif if isinstance(wif, Base58) else Base58(wif)
        
        # Decode the WIF to get raw private key bytes
        wif_bytes = unhexlify(base58_check_decode(wif))
        # Skip the prefix (0x80) and get just the private key (32 bytes)
        privkey_bytes = wif_bytes[1:33]
        
        # Create secp256k1 PrivateKey object
        self._secp256k1_privkey = secp256k1.PrivateKey(privkey_bytes, raw=True)
        # Get the corresponding public key
        self._secp256k1_pubkey = self._secp256k1_privkey.pubkey
        
        # Store compressed and uncompressed public keys
        self._pubkeyhex = self._secp256k1_pubkey.serialize(compressed=True).hex()
        self._pubkeyuncompressedhex = self._secp256k1_pubkey.serialize(compressed=False).hex()
        
        self.pubkey = PublicKey(self._pubkeyhex, prefix=prefix)
        self.uncompressed = PublicKey(self._pubkeyuncompressedhex, prefix=prefix)
        self.uncompressed.address = Address(
            pubkey=self._pubkeyuncompressedhex, prefix=prefix
        )
        self.address = Address(pubkey=self._pubkeyhex, prefix=prefix)

    def compressed_pubkey(self):
        """Derive compressed and uncompressed public keys"""
        # Just return the precomputed values
        return [self._pubkeyhex, self._pubkeyuncompressedhex]

    def __bytes__(self):
        """Return raw private key bytes"""
        return bytes(self._wif)



r"""
config.py

  ____  _ _   ____  _                         
 | __ )(_) |_/ ___|| |__   __ _ _ __ ___  ___ 
 |  _ \| | __\___ \| '_ \ / _` | '__/ _ \/ __|
 | |_) | | |_ ___) | | | | (_| | | |  __/\__ \
 |____/|_|\__|____/|_| |_|\__,_|_|  \___||___/
       ____  _             _                  
      / ___|(_) __ _ _ __ (_)_ __   __ _      
      \___ \| |/ _` | '_ \| | '_ \ / _` |     
       ___) | | (_| | | | | | | | | (_| |     
      |____/|_|\__, |_| |_|_|_| |_|\__, |     
               |___/               |___/      


WTFPL litepresence.com Dec 2021 & squidKid-deluxe Jan 2024

USER CONTROLS

TODO: Make this into a config class like that in GrapheneMetanode, so that the end
user doesn't have to schlepp around in their site-packages.
"""
import os

PATH = os.path.abspath(os.path.dirname(__file__))



# MAX is 4294967295; year 2106 due to 32 bit unsigned integer
END_OF_TIME = 4 * 10**9  # about 75 years in future
# very little
SATOSHI = 0.00000001
# almost 1
SIXSIG = 0.999999
# timeout during websocket handshake; default 4 seconds
HANDSHAKE_TIMEOUT = 4
# multiprocessing handler lifespan, default 60 seconds
PROCESS_TIMEOUT = 60
# default False for persistent limit orders
KILL_OR_FILL = False
# default True scales elements of oversize gross order to means
AUTOSCALE = True
# default True to never spend last 2 bitshares
CORE_FEES = True
# multiprocessing incarnations, default 3 attempts
ATTEMPTS = 3
# prevent extreme number of AI generated edicts; default 20
LIMIT = None
# default True to execute order in primary script process
JOIN = True
# ignore orders value less than ~X bitshares; 0 to disable
DUST = 0
# True = heavy print output
DEV = False
# Application version
VERSION = "2.0.0"
# Node list bitshares
NODES = [
    "wss://api.bts.mobi/wss",
    "wss://api-us.61bts.com/wss",
    "wss://cloud.xbts.io/ws",
    "wss://api.dex.trading/wss",
    "wss://eu.nodes.bitshares.ws/ws",
    "wss://api.pindd.club/ws",
    "wss://dex.iobanker.com/ws",
    "wss://public.xbts.io/ws",
    "wss://node.xbts.io/ws",
    "wss://node.market.rudex.org/ws",
    "wss://nexus01.co.uk/ws",
    "wss://api-bts.liondani.com/ws",
    "wss://api.bitshares.bhuz.info/wss",
    "wss://btsws.roelandp.nl/ws",
    "wss://hongkong.bitshares.im/ws",
    "wss://node1.deex.exchange/wss",
    "wss://api.cnvote.vip:888/wss",
    "wss://bts.open.icowallet.net/ws",
    "wss://api.weaccount.cn/ws",
    "wss://api.61bts.com",
    "wss://api.btsgo.net/ws",
    "wss://bitshares.bts123.cc:15138/wss",
    "wss://singapore.bitshares.im/wss",
]

# BitShares Mainnet
ID = "4018d7844c78f6a6c41c6a552b898022310fc5dec06da467ee7905a8dad512c8"
PREFIX = "BTS"
